#-----------------------------------------------------------------------------
# Copyright (c) 2005-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import ast
import os
import os.path
import sys
import py_compile
import textwrap
import zipfile

import pytest

from PyInstaller.lib.modulegraph import modulegraph
from PyInstaller.utils.tests import xfail, skipif, skipif_win, is_py2, is_py3

def _import_and_get_node(tmpdir, module_name, path=None):
    script = tmpdir.join('script.py')
    script.write('import %s' % module_name)
    if path is None:
        path = [str(tmpdir)]
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))
    return mg.findNode(module_name)


def test_sourcefile(tmpdir):
    tmpdir.join('source.py').write('###')
    node = _import_and_get_node(tmpdir, 'source')
    assert isinstance(node, modulegraph.SourceModule)


def test_invalid_sourcefile(tmpdir):
    tmpdir.join('invalid_source.py').write('invalid python-source code')
    node = _import_and_get_node(tmpdir, 'invalid_source')
    assert isinstance(node, modulegraph.InvalidSourceModule)


@skipif(is_py3, reason='Python 3 does not look into the __pycache__')
def test_compliedfile(tmpdir):
    pysrc = tmpdir.join('compiled.py')
    pysrc.write('###')
    py_compile.compile(str(pysrc))
    pysrc.remove()
    node = _import_and_get_node(tmpdir, 'compiled')
    assert isinstance(node, modulegraph.CompiledModule)


def test_invalid_compiledfile(tmpdir):
    tmpdir.join('invalid_compiled.pyc').write('invalid byte-code')
    node = _import_and_get_node(tmpdir, 'invalid_compiled')
    assert isinstance(node, modulegraph.InvalidCompiledModule)


def test_builtin(tmpdir):
    node = _import_and_get_node(tmpdir, 'sys', path=sys.path)
    assert isinstance(node, modulegraph.BuiltinModule)


def test_extension(tmpdir):
    node = _import_and_get_node(tmpdir, '_ctypes', path=sys.path)
    assert isinstance(node, modulegraph.Extension)


def test_package(tmpdir):
    pysrc = tmpdir.join('stuff', '__init__.py')
    pysrc.write('###', ensure=True)
    node = _import_and_get_node(tmpdir, 'stuff')
    assert node.__class__ is modulegraph.Package
    assert node.filename in (str(pysrc), str(pysrc)+'c')
    assert node.packagepath == [pysrc.dirname]


@skipif(is_py3, reason='Python 3 does not look into the __pycache__')
def test_compiled_package(tmpdir):
    pysrc = tmpdir.join('stuff', '__init__.py').ensure()
    pysrc.write('###')
    py_compile.compile(str(pysrc))
    pysrc.remove()
    node = _import_and_get_node(tmpdir, 'stuff')
    assert node.__class__ is modulegraph.Package
    assert node.filename == str(pysrc) + 'c'
    assert node.packagepath == [str(pysrc.dirname)]


#-- Basic tests - these seem to be missing in the original modulegraph
#-- test-suite

def test_relative_import_missing(tmpdir):
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    pkg = libdir.join('pkg')
    pkg.join('__init__.py').ensure().write('#')
    pkg.join('x', '__init__.py').ensure().write('#')
    pkg.join('x', 'y', '__init__.py').ensure().write('#')
    pkg.join('x', 'y', 'z.py').ensure().write('from . import DoesNotExist')

    script = tmpdir.join('script.py')
    script.write('import pkg.x.y.z')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))
    assert isinstance(mg.findNode('pkg.x.y.z'), modulegraph.SourceModule)
    assert isinstance(mg.findNode('pkg.x.y.DoesNotExist'),
                      modulegraph.MissingModule)


#-- Tests with a single module in a zip-file

def _zip_directory(filename, path):
    with zipfile.ZipFile(filename, mode='w') as zfh:
        for filename in path.visit(fil='*.py*'):
            zfh.write(str(filename), filename.relto(path))


def test_zipped_module_source(tmpdir):
    pysrc = tmpdir.join('stuff.py')
    pysrc.write('###', ensure=True)
    zipfilename = str(tmpdir.join('unstuff.zip'))
    _zip_directory(zipfilename, tmpdir)
    node = _import_and_get_node(tmpdir, 'stuff', path=[zipfilename])
    assert node.__class__ is modulegraph.SourceModule
    assert node.filename.startswith(os.path.join(zipfilename, 'stuff.py'))


def test_zipped_module_source_and_compiled(tmpdir):
    pysrc = tmpdir.join('stuff.py')
    pysrc.write('###', ensure=True)
    py_compile.compile(str(pysrc))
    zipfilename = str(tmpdir.join('unstuff.zip'))
    _zip_directory(zipfilename, tmpdir)
    node = _import_and_get_node(tmpdir, 'stuff', path=[zipfilename])
    # Do not care whether it's source or compiled, as long as it is
    # neither invalid nor missing.
    assert node.__class__ in (modulegraph.SourceModule, modulegraph.CompiledModule)
    assert node.filename.startswith(os.path.join(zipfilename, 'stuff.py'))


@skipif(is_py3, reason='Python 3 does not look into the __pycache__')
def test_zipped_module_compiled(tmpdir):
    pysrc = tmpdir.join('stuff.py')
    pysrc.write('###', ensure=True)
    py_compile.compile(str(pysrc))
    pysrc.remove()
    zipfilename = str(tmpdir.join('unstuff.zip'))
    _zip_directory(zipfilename, tmpdir)
    node = _import_and_get_node(tmpdir, 'stuff', path=[zipfilename])
    assert node.__class__ is modulegraph.CompiledModule
    assert node.filename.startswith(os.path.join(zipfilename, 'stuff.py'))


#-- Tests with a package in a zip-file

def _zip_package(filename, path):
    with zipfile.ZipFile(filename, mode='w') as zfh:
        for filename in path.visit():
            zfh.write(str(filename), filename.relto(path.dirname))

def test_zipped_package_source(tmpdir):
    pysrc = tmpdir.join('stuff', '__init__.py')
    pysrc.write('###', ensure=True)
    zipfilename = str(tmpdir.join('stuff.zip'))
    _zip_package(zipfilename, tmpdir.join('stuff'))
    node = _import_and_get_node(tmpdir, 'stuff', path=[zipfilename])
    assert node.__class__ is modulegraph.Package
    assert node.packagepath == [os.path.join(zipfilename, 'stuff')]


def test_zipped_package_source_and_compiled(tmpdir):
    pysrc = tmpdir.join('stuff', '__init__.py')
    pysrc.write('###', ensure=True)
    py_compile.compile(str(pysrc))
    zipfilename = str(tmpdir.join('stuff.zip'))
    _zip_package(zipfilename, tmpdir.join('stuff'))
    node = _import_and_get_node(tmpdir, 'stuff', path=[zipfilename])
    assert node.__class__ is modulegraph.Package
    assert node.packagepath == [os.path.join(zipfilename, 'stuff')]


@skipif(is_py3, reason='Python 3 does not look into the __pycache__')
def test_zipped_package_compiled(tmpdir):
    pysrc = tmpdir.join('stuff', '__init__.py')
    pysrc.write('###', ensure=True)
    py_compile.compile(str(pysrc))
    pysrc.remove()
    zipfilename = str(tmpdir.join('stuff.zip'))
    _zip_package(zipfilename, tmpdir.join('stuff'))
    node = _import_and_get_node(tmpdir, 'stuff', path=[zipfilename])
    assert node.__class__ is modulegraph.Package
    assert node.packagepath == [os.path.join(zipfilename, 'stuff')]


#-- Namespace packages

@skipif(is_py2, reason='Requires Python 3 or newer')
def test_nspackage_pep420(tmpdir):
    p1 = tmpdir.join('p1')
    p2 = tmpdir.join('p2')
    p1.join('stuff', 'a.py').ensure().write('###')
    p2.join('stuff', 'b.py').ensure().write('###')
    path = [str(p1), str(p2)]

    script = tmpdir.join('script.py')
    script.write('import stuff.a, stuff.b')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))

    mg.report()

    assert isinstance(mg.findNode('stuff.a'), modulegraph.SourceModule)
    assert isinstance(mg.findNode('stuff.b'), modulegraph.SourceModule)

    node = mg.findNode('stuff')
    assert isinstance(node, modulegraph.NamespacePackage)
    assert node.packagepath == [os.path.join(p, 'stuff') for p in path]

# :todo: test_namespace_setuptools
# :todo: test_namespace_pkg_resources

@skipif_win
def test_symlinks(tmpdir):
    base_dir = tmpdir.join('base').ensure(dir=True)
    p1_init = tmpdir.join('p1', '__init__.py').ensure()
    p2_init = tmpdir.join('p2', '__init__.py').ensure()
    p1_init.write('###')
    p2_init.write('###')

    base_dir.join('p1').ensure(dir=True)

    os.symlink(str(p1_init), str(base_dir.join('p1', '__init__.py')))
    os.symlink(str(p2_init), str(base_dir.join('p1', 'p2.py')))

    node = _import_and_get_node(base_dir, 'p1.p2')
    assert isinstance(node, modulegraph.SourceModule)


def test_import_order_1(tmpdir):
    # Ensure modulegraph processes modules in the same order as Python does.

    class MyModuleGraph(modulegraph.ModuleGraph):
        def _load_module(self, fqname, fp, pathname, info):
            if not record or record[-1] != fqname:
                record.append(fqname) # record non-consecutive entries
            return super(MyModuleGraph, self)._load_module(fqname, fp,
                                                           pathname, info)

    record = []

    for filename, content in (
        ('a/',      'from . import c, d'),
        ('a/c',            '#'),
        ('a/d/',    'from . import f, g, h'),
        ('a/d/f/',  'from . import j, k'),
        ('a/d/f/j',         '#'),
        ('a/d/f/k',         '#'),
        ('a/d/g/',   'from . import l, m'),
        ('a/d/g/l',         '#'),
        ('a/d/g/m',         '#'),
        ('a/d/h',           '#'),
        ('b/',      'from . import e'),
        ('b/e/',    'from . import i'),
        ('b/e/i',           '#')):
        if filename.endswith('/'): filename += '__init__'
        tmpdir.join(*(filename+'.py').split('/')).ensure().write(content)

    script = tmpdir.join('script.py')
    script.write('import a, b')
    mg = MyModuleGraph([str(tmpdir)])
    mg.run_script(str(script))

    # This is the order Python imports these modules given that script.
    expected = ['a',
                    'a.c', 'a.d', 'a.d.f', 'a.d.f.j', 'a.d.f.k',
                    'a.d.g', 'a.d.g.l', 'a.d.g.m',
                    'a.d.h',
               'b', 'b.e', 'b.e.i']
    assert record == expected


def test_import_order_2(tmpdir):
    # Ensure modulegraph processes modules in the same order as Python does.

    class MyModuleGraph(modulegraph.ModuleGraph):
        def _load_module(self, fqname, fp, pathname, info):
            if not record or record[-1] != fqname:
                record.append(fqname) # record non-consecutive entries
            return super(MyModuleGraph, self)._load_module(fqname, fp,
                                                           pathname, info)

    record = []

    for filename, content in (
        ('a/',      '#'),
        ('a/c/',    '#'),
        ('a/c/g',   '#'),
        ('a/c/h',   'from . import g'),
        ('a/d/',    '#'),
        ('a/d/i',   'from ..c import h'),
        ('a/d/j/',  'from .. import i'),
        ('a/d/j/o', '#'),
        ('b/',      'from .e import k'),
        ('b/e/',    'import a.c.g'),
        ('b/e/k',   'from .. import f'),
        ('b/e/l',   'import a.d.j'),
        ('b/f/',    '#'),
        ('b/f/m',   '#'),
        ('b/f/n/',  '#'),
        ('b/f/n/p', 'from ...e import l')):
        if filename.endswith('/'): filename += '__init__'
        tmpdir.join(*(filename+'.py').split('/')).ensure().write(content)

    script = tmpdir.join('script.py')
    script.write('import b.f.n.p')
    mg = MyModuleGraph([str(tmpdir)])
    mg.run_script(str(script))

    # This is the order Python imports these modules given that script.
    expected = ['b', 'b.e',
                'a', 'a.c', 'a.c.g',
                'b.e.k',
                'b.f', 'b.f.n', 'b.f.n.p',
                'b.e.l',
                'a.d', 'a.d.j', 'a.d.i', 'a.c.h']
    assert record == expected
    print(record)


#---- scan bytecode

def __scan_code(code, use_ast, monkeypatch):
    mg = modulegraph.ModuleGraph()
    # _process_imports would set _deferred_imports to None
    monkeypatch.setattr(mg, '_process_imports', lambda m: None)
    module = mg.createNode(modulegraph.Script, 'dummy.py')

    code = textwrap.dedent(code)
    if use_ast:
        co_ast = compile(code, 'dummy', 'exec', ast.PyCF_ONLY_AST)
        co = compile(co_ast, 'dummy', 'exec')
    else:
        co_ast = None
        co = compile(code, 'dummy', 'exec')
    mg._scan_code(module, co)
    return module


@pytest.mark.parametrize("use_ast", (True, False))
def test_scan_code__empty(monkeypatch, use_ast):
    code = "# empty code"
    module = __scan_code(code, use_ast, monkeypatch)
    assert len(module._deferred_imports) == 0
    assert len(module._global_attr_names) == 0


@pytest.mark.parametrize("use_ast", (True, False))
def test_scan_code__basic(monkeypatch, use_ast):
    code = """
    import os.path
    from sys import maxint, exitfunc, platform
    del exitfunc
    def testfunc():
        import shutil
    """
    module = __scan_code(code, use_ast, monkeypatch)
    assert len(module._deferred_imports) == 3
    assert ([di[1][0] for di in module._deferred_imports]
            == ['os.path', 'sys', 'shutil'])
    assert module.is_global_attr('maxint')
    assert module.is_global_attr('os')
    assert module.is_global_attr('platform')
    assert not module.is_global_attr('shutil') # not imported at module level
    assert not module.is_global_attr('exitfunc')


#-- SWIG packages - pyinstaller specific tests

def test_swig_import_simple_BUGGY(tmpdir):
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    osgeo = libdir.join('pyi_test_osgeo')
    osgeo.join('__init__.py').ensure().write('#')
    osgeo.join('pyi_gdal.py').write('# automatically generated by SWIG\n'
                                    'import _pyi_gdal')
    osgeo.join('_pyi_gdal.py').write('#')

    script = tmpdir.join('script.py')
    script.write('from pyi_test_osgeo import pyi_gdal')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))

    assert isinstance(mg.findNode('pyi_test_osgeo'), modulegraph.Package)
    assert isinstance(mg.findNode('pyi_test_osgeo.pyi_gdal'),
                      modulegraph.SourceModule)
    # The "C" module is frozen under its unqualified rather than qualified
    # name. See comment in modulegraph._safe_import_hook.
    # BUG: modulegraph contains a probable bug: Only the module's identifier
    # is changed, not the module's graphident. Thus the node is still found
    # under it's old name. The relevant code was brought from PyInstaller to
    # upstream, so this might be PyInstaller's fault. See
    # test_swig_import_simple for what it should be.
    # This is a separate test-case, not marked as xfail, so we can spot
    # whether the SWIG support works at all.
    assert isinstance(mg.findNode('pyi_test_osgeo._pyi_gdal'),
                      modulegraph.SourceModule)
    if is_py2:
        # In Python 2.7 the relative import works as expected.
        assert mg.findNode('pyi_test_osgeo._pyi_gdal').identifier \
            == 'pyi_test_osgeo._pyi_gdal'
        assert mg.findNode('_pyi_gdal') is None
    else:
        # Due the the buggy implementation, the graphident is unchanged, but
        # at least the identifier should have changed.
        assert mg.findNode('pyi_test_osgeo._pyi_gdal').identifier \
            == '_pyi_gdal'
        # Due the the buggy implementation, this node does not exist.
        assert mg.findNode('_pyi_gdal') is None
    return mg  # for use in test_swig_import_simple_BUG


@xfail
def test_swig_import_simple(tmpdir):
    # Test the expected (but not implemented) behavior if SWIG support.
    mg = test_swig_import_simple_BUGGY(tmpdir)
    # Given the bug in modulegraph (see test_swig_import_simple_BUGGY) this is
    # what would be the expected behavior.
    # TODO: When modulegraph is fixed, merge the two test-cases and correct
    # test_swig_import_from_top_level and siblings.
    assert mg.findNode('pyi_test_osgeo._pyi_gdal') is None
    assert isinstance(mg.findNode('_pyi_gdal'), modulegraph.SourceModule)


def test_swig_import_from_top_level(tmpdir):
    # While there is a SWIG wrapper module as expected, the package module
    # already imports the "C" module in the same way the SWIG wrapper would
    # do.
    # See the issue #1522 (at about 2017-04-26), pull-request #2578 and commit
    # 711e9e77c93a979a63648ba05f725b30dbb7c3cc.
    #
    # For Python > 2.6, SWIG tries to import the C module from the package's
    # directory and if this fails, uses "import _XXX" (which is the code
    # triggering import in modulegraph). For Python 2 this is a relative
    # import, but for Python 3 this is an absolute import.
    #
    # In this test-case, the package's __init__.py contains code equivalent to
    # the SWIG wrapper-module, causing the C module to be searched as an
    # absolute import (in Python 3). But the importing module is not a SWIG
    # candidate (names do not match), leading to the (absolute) C module to
    # become a MissingModule - which is okay up to this point. Now if the SWIG
    # wrapper-module imports the C module, there already is this
    # MissingModule, inhibiting modulegraph's SWIG import mechanism.
    #
    # This is where commit 711e9e77c93 steps in and tries to reimport the C
    # module (relative to the SWIG wrapper-module).
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    osgeo = libdir.join('pyi_test_osgeo')
    osgeo.join('__init__.py').ensure().write('import _pyi_gdal')
    osgeo.join('pyi_gdal.py').write('# automatically generated by SWIG\n'
                                    'import _pyi_gdal')
    osgeo.join('_pyi_gdal.py').write('#')

    script = tmpdir.join('script.py')
    script.write('from pyi_test_osgeo import pyi_gdal')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))

    assert isinstance(mg.findNode('pyi_test_osgeo'), modulegraph.Package)
    assert isinstance(mg.findNode('pyi_test_osgeo.pyi_gdal'),
                      modulegraph.SourceModule)
    # The "C" module is frozen under its unqualified rather than qualified
    # name. See comment in modulegraph._safe_import_hook.
    # Due the the buggy implementation (see test_swig_import_simple):
    assert isinstance(mg.findNode('pyi_test_osgeo._pyi_gdal'),
                      modulegraph.SourceModule)
    assert mg.findNode('_pyi_gdal') is None
    # This would be the correct implementation:
    #assert mg.findNode('pyi_test_osgeo._pyi_gdal') is None
    #assert isinstance(mg.findNode('_pyi_gdal'), modulegraph.SourceModule)


def test_swig_import_from_top_level_missing(tmpdir):
    # Like test_swig_import_from_top_level, but the "C" module is missing and
    # should be reported as a MissingModule.
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    osgeo = libdir.join('pyi_test_osgeo')
    osgeo.join('__init__.py').ensure().write('import _pyi_gdal')
    osgeo.join('pyi_gdal.py').write('# automatically generated by SWIG\n'
                                    'import _pyi_gdal')
    # no module '_pyi_gdal.py'

    script = tmpdir.join('script.py')
    script.write('from pyi_test_osgeo import pyi_gdal')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))
    assert isinstance(mg.findNode('pyi_test_osgeo'),
                      modulegraph.Package)
    assert isinstance(mg.findNode('pyi_test_osgeo.pyi_gdal'),
                      modulegraph.SourceModule)
    # BUG: Again, this is unecpected behaviour in modulegraph: While
    # MissingModule('_pyi_gdal') is (arguable) removed when trying to import
    # the SWIG C module, there is no MissingModule('pyi_test_osgeo.pyi_gdal')
    # added, but again MissingModule('_pyi_gdal'). I still need to understand
    # why.
    assert mg.findNode('pyi_test_osgeo._pyi_gdal') is None
    assert isinstance(mg.findNode('_pyi_gdal'), modulegraph.MissingModule)


def test_swig_import_from_top_level_but_nested(tmpdir):
    # Like test_swig_import_from_top_level, but both the wrapper and the "top
    # level" are nested. This is intented to test relative import of the "C"
    # module.
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    osgeo = libdir.join('pyi_test_osgeo')
    osgeo.join('__init__.py').ensure().write('#')
    osgeo.join('x', '__init__.py').ensure().write('#')
    osgeo.join('x', 'y', '__init__.py').ensure().write('import _pyi_gdal')
    osgeo.join('x', 'y', 'pyi_gdal.py').write(
        '# automatically generated by SWIG\n'
        'import _pyi_gdal')
    osgeo.join('x', 'y', '_pyi_gdal.py').write('#')

    script = tmpdir.join('script.py')
    script.write('from pyi_test_osgeo.x.y import pyi_gdal')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))

    assert isinstance(mg.findNode('pyi_test_osgeo.x.y.pyi_gdal'),
                      modulegraph.SourceModule)
    # The "C" module is frozen under its unqualified rather than qualified
    # name. See comment in modulegraph._safe_import_hook.
    # Due the the buggy implementation (see test_swig_import_simple):
    assert isinstance(mg.findNode('pyi_test_osgeo.x.y._pyi_gdal'),
                      modulegraph.SourceModule)
    assert mg.findNode('_pyi_gdal') is None
    # This would be the correct implementation:
    #assert mg.findNode('pyi_test_osgeo.x.y._pyi_gdal') is None
    #assert isinstance(mg.findNode('_pyi_gdal'), modulegraph.SourceModule)


def test_swig_top_level_but_no_swig_at_all(tmpdir):
    # From the script import an absolute module which looks like a SWIG
    # candidate but is no SWIG module. See issue #3040 ('_decimal')
    # The center of this test-case is that it doesn't raise a recursion too
    # deep error.
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    libdir.join('pyi_dezimal.py').ensure().write('import _pyi_dezimal')
    # no module '_pyi_dezimal.py'

    script = tmpdir.join('script.py')
    script.write('import pyi_dezimal')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))
    assert isinstance(mg.findNode('pyi_dezimal'), modulegraph.SourceModule)
    assert isinstance(mg.findNode('_pyi_dezimal'), modulegraph.MissingModule)


def test_swig_top_level_but_no_swig_at_all_existing(tmpdir):
    # Like test_swig_top_level_but_no_swig_at_all, but the "C" module exists.
    # The test-case is here for symmetry.
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    libdir.join('pyi_dezimal.py').ensure().write('import _pyi_dezimal')
    libdir.join('_pyi_dezimal.py').ensure().write('#')

    script = tmpdir.join('script.py')
    script.write('import pyi_dezimal')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))
    assert isinstance(mg.findNode('pyi_dezimal'), modulegraph.SourceModule)
    assert isinstance(mg.findNode('_pyi_dezimal'), modulegraph.SourceModule)


def test_swig_candidate_but_not_swig(tmpdir):
    # From a package module import an absolute module which looks like a SWIG
    # candidate but is no SWIG module . See issue #2911 (tifffile).
    # The center of this test-case is that it doesn't raise a recursion too
    # deep error.
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    pkg = libdir.join('pkg')
    pkg.join('__init__.py').ensure().write('from . import mymod')
    pkg.join('mymod.py').write('import _mymod')
    pkg.join('_mymod.py').write('#')

    script = tmpdir.join('script.py')
    script.write('from pkg import XXX')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))
    assert isinstance(mg.findNode('pkg'), modulegraph.Package)
    assert isinstance(mg.findNode('pkg.mymod'), modulegraph.SourceModule)
    if is_py2:
        # In Python 2 this is a relative import, global module should exist
        assert isinstance(mg.findNode('pkg._mymod'), modulegraph.SourceModule)
        assert mg.findNode('_mymod') is None
    else:
        assert mg.findNode('pkg._mymod') is None
        # This is not a SWIG module, thus the SWIG import mechanism should not
        # trigger.
        assert isinstance(mg.findNode('_mymod'), modulegraph.MissingModule)


def test_swig_candidate_but_not_swig2(tmpdir):
    """
    Variation of test_swig_candidate_but_not_swig using differnt import
    statements (like tifffile/tifffile.py does)
    """
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    pkg = libdir.join('pkg')
    pkg.join('__init__.py').ensure().write('from . import mymod')
    pkg.join('mymod.py').write('from . import _mymod\n'
                               'import _mymod')
    pkg.join('_mymod.py').write('#')

    script = tmpdir.join('script.py')
    script.write('from pkg import XXX')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))
    assert isinstance(mg.findNode('pkg'), modulegraph.Package)
    assert isinstance(mg.findNode('pkg.mymod'), modulegraph.SourceModule)
    assert isinstance(mg.findNode('pkg._mymod'), modulegraph.SourceModule)
    if is_py2:
        # In Python 2 both are relative imports, global module should not exist
        assert mg.findNode('_mymod') is None
    else:
        assert isinstance(mg.findNode('_mymod'), modulegraph.MissingModule)


def test_swig_candidate_but_not_swig_missing(tmpdir):
    # Like test_swig_candidate_but_not_swig, but the "C" module is missing and
    # should be reported as a MissingModule.
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    pkg = libdir.join('pkg')
    pkg.join('__init__.py').ensure().write('from . import mymod')
    pkg.join('mymod.py').write('import _mymod')
    # no module '_mymod.py'

    script = tmpdir.join('script.py')
    script.write('import pkg')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))
    assert isinstance(mg.findNode('pkg'), modulegraph.Package)
    assert isinstance(mg.findNode('pkg.mymod'), modulegraph.SourceModule)
    assert mg.findNode('pkg._mymod') is None
    assert isinstance(mg.findNode('_mymod'), modulegraph.MissingModule)


def test_swig_candidate_but_not_swig_missing2(tmpdir):
    """
    Variation of test_swig_candidate_but_not_swig_missing using differnt import
    statements (like tifffile/tifffile.py does)
    """
    libdir = tmpdir.join('lib')
    path = [str(libdir)]
    pkg = libdir.join('pkg')
    pkg.join('__init__.py').ensure().write('from . import mymod')
    pkg.join('mymod.py').write('from . import _mymod\n'
                               'import _mymod')
    # no module '_mymod.py'

    script = tmpdir.join('script.py')
    script.write('import pkg')
    mg = modulegraph.ModuleGraph(path)
    mg.run_script(str(script))
    assert isinstance(mg.findNode('pkg'), modulegraph.Package)
    assert isinstance(mg.findNode('pkg.mymod'), modulegraph.SourceModule)
    assert isinstance(mg.findNode('pkg._mymod'), modulegraph.MissingModule)
    if is_py2:
        # In Python 2 both are relative imports, global module should not exist
        assert mg.findNode('_mymod') is None
    else:
        assert isinstance(mg.findNode('_mymod'), modulegraph.MissingModule)



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import pkgutil
import pkg3

expected_data = 'This is data text for testing the packaging module data.'.encode('ascii')

data = pkgutil.get_data('pkg3', 'sample-data.txt')
if not data:
    raise SystemExit('Error: Could not read data with pkgutil.get_data().')

if data.strip() != expected_data:
    raise SystemExit('Error: Read data is wrong: %r' % data)
print('Okay: Resource data read.')



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


import os
import sys


if __name__ == '__main__':

    filename =  os.path.join(os.path.dirname(sys.executable),
            'test_onefile_nestedlaunch0.exe')

    try:
        import subprocess
    except ImportError:
        if os.system(filename) != 0:
            raise RuntimeError("os.system failed: %s" % filename)
    else:
        subprocess.check_call([filename])



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

"""
Functional tests for the Django content management system (CMS).
"""

import pytest

from PyInstaller.utils.tests import importorskip, skipif
from PyInstaller.compat import is_py35, is_py36


# In Django 2.1, ``django/contrib/auth/password_validation.py``, line 168, which
# is
#
#   ``DEFAULT_PASSWORD_LIST_PATH = Path(__file__).resolve().parent / 'common-passwords.txt.gz'``,
#
# the call to ``resolve()`` causes Python 3.5 to raise an exception that
# ``password_validation.pyc`` doesn't exist. Python 3.6 added the default
# argument ``strict=False``, which ignores this exception. This file is in the
# archive, but not the filesystem.
@skipif(is_py35 and not is_py36, reason='Call to resolve() raises an exception '
        'in Python 3.5.')
@importorskip('django')
# Django test might sometimes hang.
@pytest.mark.timeout(timeout=7*60)
def test_django(pyi_builder, monkeypatch, data_dir):
    # Extend sys.path so PyInstaller could find modules from 'tmpdir/django/'.
    monkeypatch.syspath_prepend(data_dir.strpath)
    # Django uses manage.py as the main script.
    script = str(data_dir / 'manage.py')
    # Create the exe, run django command 'check' to do basic sanity
    # checking of the executable.
    pyi_builder.test_script(script, app_name='django_site', app_args=['check'])



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


"""
Interactive tests are successful when they are able to run
the executable for some time. Otherwise it is marked as fail.

Note: All tests in this file should use the argument 'runtime'.
"""
import pytest

from PyInstaller.utils.tests import importorskip, xfail
from PyInstaller.compat import is_win

_RUNTIME = 10  # In seconds.


@importorskip('IPython')
@pytest.mark.skipif(is_win, reason='See issue #3535.')
def test_ipython(pyi_builder):
    pyi_builder.test_source(
        """
        from IPython import embed
        embed()
        """, runtime=_RUNTIME)


@xfail(reason='TODO - known to fail')
@importorskip('PySide')
def test_pyside(pyi_builder):
    pyi_builder.test_script('pyi_interact_pyside.py', #pyi_args=['--windowed'],
                            runtime=_RUNTIME)



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


from threading import Thread

from PyInstaller.compat import is_py2
from PyInstaller.loader.pyimod02_archive import ArchiveFile

if is_py2:
    from Queue import Queue
else:
    from queue import Queue


def test_threading_import(tmpdir):
    """
    On Python 3.3+, PyInstaller doesn't acquire a lock when performing an
    import. Therefore, two thread could both be reading the .pyz archive at the
    same time. At the core, the ArchiveFile class performs these reads. This
    test verifies that multi-threaded reads work.

    For more information, see https://github.com/pyinstaller/pyinstaller/pull/2010.
    """

    # Create a temporary file and use the ArchiveReader on it.
    tmp_file = tmpdir.join('test.txt')
    tmp_file.write('Testing')
    ar = ArchiveFile(tmp_file.strpath, 'r')

    # Use queues to synchronize threads.
    q1 = Queue()
    q2 = Queue()

    # This function, which is run in a separate thread, works to ensure that
    # both threads open a file at the same time.
    def foo():
        with ar:
            # Wait until both threads have opened the file.
            q1.put(1)
            assert q2.get() == 2
            # Wait until the other thread has closed the file before closing it
            # here.
            assert q2.get() == 3

    thread = Thread(target=foo)
    thread.start()

    # This code works with ``foo`` above to open the same file from two threads.
    with ar:
        # Wait until both threads have opened the file.
        q2.put(2)
        assert q1.get() == 1
    # Make the other thread wait until this thread has closed the file.
    q2.put(3)

    # Wait for the other thread to finish.
    thread.join()




#-----------------------------------------------------------------------------
# Copyright (c) 2013-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


# Install necessary 3rd party Python modules to run all tests.
# This script is supposed to be used in a continuous integration system:
# https://jenkins.shiningpanda.com/pyinstaller/
# Python there is mostly 64bit. Only Python 2.4 is 32bit on Windows 7.

from __future__ import print_function

import glob
import optparse
import os
import platform
import sys

# easy_install command used in a Python script.
from setuptools.command import easy_install


# Expand PYTHONPATH with PyInstaller package to support running without
# installation -- only if not running in a virtualenv.
if not hasattr(sys, 'real_prefix'):
    pyi_home = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')
    sys.path.insert(0, pyi_home)


from PyInstaller.compat import is_py2


PYVER = '.'.join([str(x) for x in sys.version_info[0:2]])


def py_arch():
    """
    .exe installers of Python modules contain architecture name in filename.
    """
    mapping = {'32bit': 'win32', '64bit': 'win-amd64'}
    arch = platform.architecture()[0]
    return mapping[arch]


# Pick a PyCrypto version to use to test bytecode encryption. This is typically
# used in conjunction with continuous integration.
if 'PYCRYPTO_VERSION' in os.environ:
    pycrypto = 'PyCrypto==%s' % os.environ['PYCRYPTO_VERSION']
else:
    pycrypto = 'PyCrypto'


_PACKAGES = {
    # 'modulename': 'pypi_name_or_url_or_path'
    'docutils': ['docutils'],
    'IPython': ['IPython'],
    'jinja2': ['jinja2'],
    'keyring': ['keyring'],
    'markdown': ['markdown'],
    'MySQLdb': ['MySQL-python-*%s-py%s.exe' % (py_arch(), PYVER)],
    'numpy': ['numpy-unoptimized-*%s-py%s.exe' % (py_arch(), PYVER)],
    'PIL': ['PIL-*%s-py%s.exe' % (py_arch(), PYVER)],
    'pycparser': ['pycparser'],
    'PyCrypto': [pycrypto],
    'psycopg2': ['psycopg2-*%s-py%s.exe' % (py_arch(), PYVER)],
    'pyexcelerate': ['pyexcelerate'],
    'pylint': ['pylint'],
    'pyodbc': ['pyodbc'],
    'pyttsx': ['pyttsx'],
    'pytz': ['pytz'],
    'pyusb': ['pyusb'], # package usb
    'qt4reactor': ['qt4reactor'],
    'requests': ['requests'],
    'scapy': ['scapy'],
    #'simplejson': ['simplejson'],
    'sphinx': ['sphinx'],
    'sqlalchemy': ['SQLAlchemy-*%s-py%s.exe' % (py_arch(), PYVER)],
    'twisted': ['twisted'],
    'wx': ['wxPython-common-*%s-py%s.exe' % (py_arch(), PYVER),
        'wxPython-2*%s-py%s.exe' % (py_arch(), PYVER)],
    # PyWin32 is installed on ShiningPanda hosting.
    'win32api': ['http://downloads.sourceforge.net/project/pywin32/pywin32/Build%%20217/pywin32-217.%s-py%s.exe' %
        (py_arch(), PYVER)],
}

_PY_VERSION = {
    'keyring': is_py2,
    'MySQLdb': is_py2,
    'numpy': is_py2,
    'PIL': is_py2,
    'psycopg2': is_py2,
    'simplejson': is_py2,
    'twisted': is_py2,  # wheels are avalable for 2.7 only
    # Installers are available only for Python 2.6/2.7.
    'wx': is_py2,
}


def main():
    parser = optparse.OptionParser()
    parser.add_option('-d', '--download-dir',
        help='Directory with maually downloaded python modules.'
    )
    opts, _ = parser.parse_args()

    # Install packages.
    for k, v in _PACKAGES.items():
        # Test python version for module.
        if k in _PY_VERSION:
            # Python version too old, skip module.
            if PYVER < _PY_VERSION[k]:
                continue
        try:
            __import__(k)
            print('Already installed...', k)
        # Module is not available - install it.
        except ImportError:
            # If not url or module name then look for installers in download area.
            if not v[0].startswith('http') and v[0].endswith('exe'):
                files = []
                # Try all file patterns.
                for pattern in v:
                    pattern = os.path.join(opts.download_dir, pattern)
                    files += glob.glob(pattern)
                # No file with that pattern was not found - skip it.
                if not files:
                    print('Skipping module...', k)
                    continue
                # Full path to installers in download directory.
                v = files
            print('Installing module...', k)
            # Some modules might require several .exe files to install.
            for f in v:
                print(' ', f)
                # Use --no-deps ... installing module dependencies might fail
                # because easy_install tries to install the same module from
                # PYPI from source code and if fails because of C code that
                # that needs to be compiled.
                try:
                    easy_install.main(['--no-deps', '--always-unzip', f])
                except Exception:
                    print(' ', k, 'installation failed')


if __name__ == '__main__':
    main()



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

# With module 'pkg_resources' it should not matter if a file is stored
# on file system, in zip archive or bundled with frozen app.
import pkg_resources as res
import pkg3

expected_data = 'This is data text for testing the packaging module data.'.encode('ascii')

# With frozen app the resources is available in directory
# os.path.join(sys._MEIPASS, 'pkg3/sample-data.txt')
data = res.resource_string(pkg3.__name__, 'sample-data.txt')
if not data:
    raise SystemExit('Error: Could not read data with pkgutil.get_data().')

if data.strip() != expected_data:
    raise SystemExit('Error: Read data is wrong: %r' % data)
print('Okay: Resource data read.')



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from __future__ import print_function

# In dist directory are Python C-extension file names like module.submodule.so
# E.g.  ./simplejson/_speedups.so  ->  ./simplejson._speedups.so

import os
import sys

# List of suffixes for Python C extension modules.
try:
    # In Python 3.3+ There is a list
    from importlib.machinery import EXTENSION_SUFFIXES
except ImportError:
    # Python 2 does not have this
    import imp
    EXTENSION_SUFFIXES = [f[0] for f in imp.get_suffixes()
                          if f[2] == imp.C_EXTENSION]

from simplejson import _speedups

modpath = os.path.join(sys.prefix, 'simplejson._speedups')
frozen_modpath = _speedups.__file__

print('Module path expected:', modpath, '+ ext', file=sys.stderr)
print('Module path  current:', frozen_modpath, file=sys.stderr)

# In Python3, filename extensions can include several dots (e.g.
# '.cpython-33m.so'), so we can not simply use os.path.splitext(), but
# have to loop over all possible extensions.
for ext in EXTENSION_SUFFIXES:
    if modpath + ext == frozen_modpath:
        print('             matched:', modpath + ext, file=sys.stderr)
        break
else:
    if not getattr(sys, 'frozen', False):
        raise SystemExit('This script only works corretly when frozen')
    raise SystemExit('Python C-extension file name is not correct.')



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


import numpy
from matplotlib import mlab
from matplotlib import pyplot


def main():
    # Part of the example at 
    # http://matplotlib.sourceforge.net/plot_directive/mpl_examples/pylab_examples/contour_demo.py
    delta = 0.025
    x = numpy.arange(-3.0, 3.0, delta)
    y = numpy.arange(-2.0, 2.0, delta)
    X, Y = numpy.meshgrid(x, y)
    Z1 = mlab.bivariate_normal(X, Y, 1.0, 1.0, 0.0, 0.0)
    Z2 = mlab.bivariate_normal(X, Y, 1.5, 0.5, 1, 1)
    Z = 10.0 * (Z2 - Z1)
    pyplot.figure()
    CS = pyplot.contour(X, Y, Z)
    pyplot.show()


if __name__ == "__main__":
    main()




hiddenimports = ['pkg_without_hook_for_pkg.sub1.sub11']

NAME = 'pkg_without_hook_for_pkg.sub1'

# self-test for this test-case:
import PyInstaller.hooks
import os
# TODO For Python 3 replace this by importlib, imp is deprecated
import imp

# 1. ensure self-test is working by searching for _this_ hook
hookmodnm = 'hook-' + NAME
searchpath = PyInstaller.hooks.__path__ + [os.path.dirname(__file__)]
assert imp.find_module(hookmodnm, searchpath) is not None, "Error in the hook's self-test"

# 2. The actual self-test: there must be no hook for the parent module
hookmodnm = 'hook-pkg_without_hook_for_pkg'
try:
    imp.find_module(hookmodnm, searchpath)
    raise Exception('Self-test of hook %s failed: hook for parent exists'
                    % NAME)
except ImportError as e:
    if (hasattr(e, 'name') and e.name == hookmodnm): # python3
        pass # okay
    elif  not e.args[0].endswith(' '+hookmodnm): # python2
        raise Exception('Self-test of hook %s failed: hook for parent exists '
                        'and has import errors.' % NAME)



#!/usr/bin/env python
'''
This script reformats the output of `myprog --help` to decent rst.

Currently handles optparse and argparse output. Removes everything
before "Options" resp. "optional arguments" and after "Obsolete
options".
'''

from __future__ import print_function

import argparse
import subprocess
import textwrap
import re
import sys
import os

__copyright__ = "Copyright (c) 2015-2018 PyInstaller Development Team, Copyright (c) 2015-2018 Hartmut Goebel"
__author__ = "Hartmut Goebel <h.goebel@crazy-compilers.com>"


def gen_headings(program, text, headings_character):
    text = (t.rstrip() for t in text.splitlines())
    new_text = []
    dedent = 0
    for line in text:
        heading = line and not line.startswith((' ', '-'))
        if heading:
            # remove trailing colon
            line = line.rstrip(':')
            dedent = 0
        elif dedent:
            # dedent for this section already known
            assert line[:dedent].strip() == '', line
            line = line[dedent:]
        else:
            # dedent for this section not yet known
            # check for next optional arguments (starting with a dash)
            leading_whitespace = re.findall(r'^[ \t]*(?=-)', line)
            if leading_whitespace:
                dedent = len(leading_whitespace[0])
                line = line[dedent:]
        if heading:
            # add a sphinx reference label
            new_text.extend(('',
                             '.. _%s %s:' % (program, line.lower()),
                             ''))
        new_text.append(line)
        if heading:
            # append underline
            new_text.append(headings_character * len(line))
            new_text.append('')  # empty line for readability only
    return '\n'.join(new_text)


def process(program, generate_headings, headings_character):
    help = subprocess.check_output([sys.executable, program, '--help'],
                                   universal_newlines=True)
    if '\nOptions:' in help:
        # optparse style
        help = help.split('\nOptions:', 1)[1]
    elif '\noptional arguments:' in help:
        # argparse style
        help = help.split('\noptional arguments:', 1)[1]
    else:
        raise SystemError('unexpected format of output for --help')
    # Remove any obsolete options
    help = re.split(r'\n\s*Obsolete options', help)[0]

    help = help.strip('\n')
    # escape stars prior to other processing
    help = help.replace('*', r'\*')
    if generate_headings:
        program = os.path.splitext(os.path.basename(program))[0].lower()
        help = textwrap.dedent(help)
        help = gen_headings(program, help, headings_character)
    return help


def to_file(program, generate_headings, headings_character, outfile):
    help = process(program, generate_headings, headings_character)
    with open(outfile, 'w') as ofh:
        print(help, file=ofh)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-headings',
                        action='store_true', default=True,
                        help=("Generate section headings from arumgent groups"
                              " (this is the default)"))
    parser.add_argument('--no-generate-headings',action='store_false',
                        dest='generate_headings')
    g = parser.add_argument_group('tst')
    g.add_argument('--headings-character', default="-",
                        help=("Character to use for underlining section headers"
                              " (default: -)"))
    parser.add_argument('program')
    args = parser.parse_args()
    print(process(**vars(args)))

if __name__ == '__main__':
    main()



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


from PyInstaller.depend import analysis
from PyInstaller.building.build_main import Analysis

def test_issue_2492(monkeypatch, tmpdir):
    # Crash if an extension module has an hidden import to ctypes (e.g. added
    # by the hook).

    # Need to set up some values
    monkeypatch.setattr('PyInstaller.config.CONF',
                        {'workpath': str(tmpdir),
                         'spec': str(tmpdir),
                         'warnfile': str(tmpdir.join('warn.txt')),
                         'dot-file': str(tmpdir.join('imports.dot')),
                         'xref-file': str(tmpdir.join('imports.xref')),
                         'hiddenimports': [],
                         'specnm': 'issue_2492_script'})
    # Speedup: avoid analyzing base_library.zip
    monkeypatch.setattr(analysis, 'PY3_BASE_MODULES', [])

    script = tmpdir.join('script.py')
    script.write('import _struct')
    # create a hook
    tmpdir.join('hook-_struct.py').write('hiddenimports = ["ctypes"]')
    a = Analysis([str(script)], hookspath=[str(tmpdir)],
                 excludes=['encodings', 'pydoc', 'xml', 'distutils'])



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import multiprocessing


def f(x):
    return x*x


if __name__ == '__main__':
    multiprocessing.freeze_support()
    # Start 4 worker processes.
    pool = multiprocessing.Pool(processes=4)
    print('Evaluate "f(10)" asynchronously.')
    res = pool.apply_async(f, [10])
    print(res.get(timeout=1))          # prints "100"
    print('Print "[0, 1, 4,..., 81]"')
    print(pool.map(f, range(10)))



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

"""
Functional test exercising the non-default protocol `arg1` of version 3 of the
PyPubSub API.
"""

from wx.lib.pubsub import setuparg1
from wx.lib.pubsub import pub as Publisher

def on_message(message):
    print('Message received.')
    if not message.data == 762:
        raise SystemExit(
            'Message data "762" expected but received "%s".' % str(message.data))

Publisher.subscribe(on_message, 'topic.subtopic')
Publisher.sendMessage('topic.subtopic', 762)



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


import os
import pytest
import shutil
from os.path import join

from PyInstaller.utils.hooks import collect_data_files, collect_submodules, \
    get_module_file_attribute, remove_prefix, remove_suffix, \
    remove_file_extension, is_module_or_submodule, \
    is_module_satisfies
from PyInstaller.compat import exec_python, ALL_SUFFIXES


class TestRemovePrefix(object):
    # Verify that removing a prefix from an empty string is OK.
    def test_empty_string(self):
        assert '' == remove_prefix('', 'prefix')

    # An empty prefix should pass the string through unmodified.
    def test_emptystr_unmodif(self):
        assert 'test' == remove_prefix('test', '')

    # If the string is the prefix, it should be empty at exit.
    def test_string_prefix(self):
        assert '' == remove_prefix('test', 'test')

    # Just the prefix should be removed.
    def test_just_prefix(self):
        assert 'ing' == remove_prefix('testing', 'test')

    # A matching string not as prefix should produce no modifications
    def test_no_modific(self):
        assert 'atest' == remove_prefix('atest', 'test')


class TestRemoveSuffix(object):
    # Verify that removing a suffix from an empty string is OK.
    def test_empty_string(self):
        assert '' == remove_suffix('', 'suffix')

    # An empty suffix should pass the string through unmodified.
    def test_emptystr_unmodif(self):
        assert 'test' == remove_suffix('test', '')

    # If the string is the suffix, it should be empty at exit.
    def test_string_suffix(self):
        assert '' == remove_suffix('test', 'test')

    # Just the suffix should be removed.
    def test_just_suffix(self):
        assert 'test' == remove_suffix('testing', 'ing')

    # A matching string not as suffix should produce no modifications
    def test_no_modific(self):
        assert 'testa' == remove_suffix('testa', 'test')


class TestRemoveExtension(object):
    # Removing a suffix from a filename with no extension returns the filename.
    def test_no_extension(self):
        assert 'file' == remove_file_extension('file')

    # A filename with two extensions should have only the first removed.
    def test_two_extensions(self):
        assert 'file.1' == remove_file_extension('file.1.2')

    # Standard case - remove an extension
    def test_remove_ext(self):
        assert 'file' == remove_file_extension('file.1')

    # Unix-style .files are not treated as extensions
    def test_unixstyle_not_ext(self):
        assert '.file' == remove_file_extension('.file')

    # Unix-style .file.ext works
    def test_unixstyle_ext(self):
        assert '.file' == remove_file_extension('.file.1')

    # Unix-style .file.ext works
    def test_unixstyle_path(self):
        assert '/a/b/c' == remove_file_extension('/a/b/c')
        assert '/a/b/c' == remove_file_extension('/a/b/c.1')

    # Windows-style .file.ext works
    def test_win32style_path(self):
        assert 'C:\\a\\b\\c' == remove_file_extension('C:\\a\\b\\c')
        assert 'C:\\a\\b\\c' == remove_file_extension('C:\\a\\b\\c.1')


# The name of the hookutils test files directory
TEST_MOD = 'hookutils_package'
# The path to this directory.
TEST_MOD_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'hookutils_files')


@pytest.fixture
def mod_list(monkeypatch):
    # Add 'hookutils_files' to sys.path (so ``is_package`` can find it) and to
    # ``pathex`` (so code run in a subprocess can find it).
    monkeypatch.setattr('PyInstaller.config.CONF', {'pathex': [TEST_MOD_PATH]})
    monkeypatch.syspath_prepend(TEST_MOD_PATH)
    # Use the hookutils_test_files package for testing.
    return collect_submodules(TEST_MOD)

class TestCollectSubmodules(object):
    # An error should be thrown if a module, not a package, was passed.
    def test_collect_submod_module(self):
        # os is a module, not a package.
        with pytest.raises(ValueError):
            collect_submodules(__import__('os'))

    # The package name itself should be in the returned list.
    def test_collect_submod_itself(self, mod_list):
        assert TEST_MOD in mod_list

    # Python extension is included in the list.
    def test_collect_submod_pyextension(self, mod_list):
        assert TEST_MOD + '.pyextension' in mod_list

    # Check that all packages get included
    def test_collect_submod_all_included(self, mod_list):
        mod_list.sort()
        assert mod_list == [TEST_MOD,
                            # Python extensions on Windows ends with '.pyd' and
                            # '.so' on Linux, Mac OS X and other operating systems.
                            TEST_MOD + '.pyextension',
                            TEST_MOD + '.subpkg',
                            TEST_MOD + '.subpkg.twelve',
                            TEST_MOD + '.two']

    # Dynamic libraries (.dll, .dylib) are not included in the list.
    def test_collect_submod_no_dynamiclib(self, mod_list):
        assert TEST_MOD + '.dynamiclib' not in mod_list

    # Subpackages without an __init__.py should not be included.
    def test_collect_submod_subpkg_init(self, mod_list):
        assert TEST_MOD + '.py_files_not_in_package.sub_pkg.three' not in mod_list

    # Test with a subpackage.
    def test_collect_submod_subpkg(self, mod_list):
        # Note: Even though mod_list is overwritten, it's still needed as a
        # fixture, so that the path to the TEST_MOD will be set correctly.
        mod_list = collect_submodules(TEST_MOD + '.subpkg')
        mod_list.sort()
        assert mod_list == [TEST_MOD + '.subpkg',
                            TEST_MOD + '.subpkg.twelve']

    # Test in an ``.egg`` file.
    def test_collect_submod_egg(self, tmpdir, monkeypatch):
        # Copy files to a tmpdir for egg building.
        dest_path = tmpdir.join('hookutils_package')
        shutil.copytree(TEST_MOD_PATH, dest_path.strpath)
        monkeypatch.chdir(dest_path)

        # Create an egg from the test package. For debug, show the output of
        # the egg build.
        print(exec_python('setup.py', 'bdist_egg'))

        # Obtain the name of the egg, which depends on the Python version.
        dist_path = dest_path.join('dist')
        fl = os.listdir(dist_path.strpath)
        assert len(fl) == 1
        egg_name = fl[0]
        assert egg_name.endswith('.egg')

        # Add the egg to Python's path.
        pth = dist_path.join(egg_name).strpath
        monkeypatch.setattr('PyInstaller.config.CONF', {'pathex': [pth]})
        monkeypatch.syspath_prepend(pth)

        # Verify its contents.
        ml = collect_submodules(TEST_MOD)
        self.test_collect_submod_all_included(ml)


def test_is_module_or_submodule():
    assert is_module_or_submodule('foo.bar', 'foo.bar')
    assert is_module_or_submodule('foo.bar.baz', 'foo.bar')
    assert not is_module_or_submodule('foo.bard', 'foo.bar')
    assert not is_module_or_submodule('foo', 'foo.bar')


def test_is_module_satisfies_package_not_installed():
    assert is_module_satisfies('pytest')
    assert not is_module_satisfies('magnumopus-no-package-test-case')


_DATA_BASEPATH = join(TEST_MOD_PATH, TEST_MOD)
_DATA_PARAMS = [
    (TEST_MOD, ('dynamiclib.dll',
                'dynamiclib.dylib',
                'nine.dat',
                join('py_files_not_in_package', 'data', 'eleven.dat'),
                join('py_files_not_in_package', 'ten.dat'),
                'pyextension.pyd',
                'pyextension.so',
                join('subpkg', 'thirteen.txt'),
    )),
    (TEST_MOD + '.subpkg', (
                join('subpkg', 'thirteen.txt'),
    ))
]


@pytest.fixture(params=_DATA_PARAMS, ids=['package', 'subpackage'])
def data_lists(monkeypatch, request):
    def _sort(sequence):
        l = list(sequence)
        l.sort()
        return tuple(l)
    # Add path with 'hookutils_files' module to ``sys.path`` so tests
    # could find this module - useful for subprocesses.
    monkeypatch.syspath_prepend(TEST_MOD_PATH)
    # Use the hookutils_test_files package for testing.
    mod_name = request.param[0]
    data = collect_data_files(mod_name)
    # Break list of (source, dest) into source and dest lists.
    subfiles = request.param[1]
    src = [item[0] for item in data]
    dst = [item[1] for item in data]

    return subfiles, _sort(src), _sort(dst)


# An error should be thrown if a module, not a package, was passed.
def test_collect_data_module():
    # 'os' is a module, not a package.
    with pytest.raises(ValueError):
        collect_data_files(__import__('os'))


# Make sure only data files are found.
def test_collect_data_no_extensions(data_lists):
    subfiles, src, dst = data_lists
    for item in ['pyextension.pyd', 'pyextension.so']:
        # Only text valid extensions for the current platform.
        if os.path.splitext(item)[1] in ALL_SUFFIXES:
            item = join(_DATA_BASEPATH, item)
            print(src)
            assert item not in src


# Make sure all data files are found.
def test_collect_data_all_included(data_lists):
    subfiles, src, dst = data_lists
    # Check the source and dest lists against the correct values in
    # subfiles.
    print(subfiles)
    src_compare = tuple([join(_DATA_BASEPATH, subpath) for subpath in subfiles
                         if os.path.splitext(subpath)[1] not in ALL_SUFFIXES])
    dst_compare = [os.path.dirname(join(TEST_MOD, subpath))
                   for subpath in subfiles
                   if os.path.splitext(subpath)[1] not in ALL_SUFFIXES]
    dst_compare.sort()
    dst_compare = tuple(dst_compare)
    assert src == src_compare
    assert dst == dst_compare


# An Import error should be thrown if a module is not found.
def test_get_module_file_attribute_non_exist_module():
    with pytest.raises(ImportError):
        get_module_file_attribute('pyinst_nonexisting_module_name')



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2018, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License with exception
# for distributing bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


import sys

from PyQt5 import Qt
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets


def main():
    app = QtWidgets.QApplication(sys.argv)
    read_formats = ', '.join([str(format).lower() \
        for format in QtGui.QImageReader.supportedImageFormats()])
    print(("Qt5 plugin paths: " + str(list(app.libraryPaths()))))
    print(("Qt5 image read support: " + read_formats))
    print(('Qt5 Libraries path: ' + \
           str(QtCore.QLibraryInfo.location(QtCore.QLibraryInfo.LibrariesPath))))
    label = QtWidgets.QLabel("Hello World from PyQt5", None)
    label.setWindowTitle("Hello World from PyQt5")
    label.resize(300, 300)
    label.show()
    app.exec_()


if __name__ == "__main__":
    main()
